-module(tpTracerFile).
-include("eTpf.hrl").

-export([
   start_link/1
   , init/2
]).

%% sys callbacks
-export([
   system_continue/3
   , system_terminate/4
   , system_code_change/4
]).

-record(state, {
   parent :: pid()
   , fBaseName :: file:filename_all()
   , size = 0 :: non_neg_integer()
   , fMaxSize :: infinity | non_neg_integer()
   , ioDevice :: file:io_device()
   , fEventFrame :: pos_integer()
   , eventsThisFrame = 0 :: non_neg_integer()
   , buffer = <<>> :: binary()

   , tcmIsCut = false
}).

start_link(Opts) ->
   Pid = proc_lib:spawn_link(?MODULE, init, [self(), Opts]),
   {ok, Pid}.

init(Parent, TracerOpts) ->
   process_flag(message_queue_data, off_heap),
   process_flag(trap_exit, true),
   %% No need to close the file, it'll be closed when the process exits.
   Filename = maps:get(fBaseName, TracerOpts, "traces.lz4"),
   {ok, IoDevice} = file:open(Filename, [write, raw]),

   TcmIsCut = maps:get(tcmIsCut, TracerOpts, maps:get(tcmDepth, ?defTcmMap)),
   case TcmIsCut of
      true ->
         erlang:put(tcmDepth, maps:get(tcmDepth, TracerOpts, maps:get(tcmDepth, ?defTcmMap))),
         erlang:put(tcmListSize, maps:get(tcmListSize, TracerOpts, maps:get(tcmListSize, ?defTcmMap))),
         erlang:put(tcmMapSize, maps:get(tcmMapSize, TracerOpts, maps:get(tcmMapSize, ?defTcmMap))),
         erlang:put(tcmTupleSize, maps:get(tcmTupleSize, TracerOpts, maps:get(tcmTupleSize, ?defTcmMap))),
         erlang:put(tcmBinSize, maps:get(tcmBinSize, TracerOpts, maps:get(tcmBinSize, ?defTcmMap))),
         erlang:put(tcmBitSize, maps:get(tcmBitSize, TracerOpts, maps:get(tcmBitSize, ?defTcmMap))),
         erlang:put(tcmNestStruct, maps:get(tcmNestStruct, TracerOpts, maps:get(tcmNestStruct, ?defTcmMap)));
      _ ->
         ignore
   end,

   State = #state{
      parent = Parent
      , fBaseName = Filename
      , ioDevice = IoDevice
      , fMaxSize = maps:get(fMaxSize, TracerOpts, infinity)
      , fEventFrame = maps:get(fEventFrame, TracerOpts, 100000)
      , tcmIsCut = TcmIsCut
   },
   loop(State).

loop(State = #state{parent = Parent, size = Size, ioDevice = IoDevice, fEventFrame = MaxEvents, eventsThisFrame = NumEvents0, buffer = Buffer0, tcmIsCut = TcmIsCut}) ->
   receive
      {system, From, Request} ->
         sys:handle_system_msg(Request, From, Parent, ?MODULE, [], State);
      {'EXIT', Parent, Reason} ->
         terminate(Reason, State);
      RMsg ->
         Msg = case TcmIsCut of true -> tpTermCut:cut(RMsg); _ -> RMsg end,
         Bin = term_to_binary(Msg),
         BinSize = byte_size(Bin),
         Buffer = <<Buffer0/binary, BinSize:32, Bin/binary>>,
         NumEvents = NumEvents0 + 1,
         if
            MaxEvents =:= NumEvents ->
               Frame = lz4f:compress_frame(Buffer),
               ok = file:write(IoDevice, Frame),
               maybe_rotate(State#state{size = Size + byte_size(Frame),
                  eventsThisFrame = 0, buffer = <<>>});
            true ->
               loop(State#state{eventsThisFrame = NumEvents, buffer = Buffer})
         end
   end.

%%IMY-todo 干掉这个函数 同时对于旋转的文件名 使用时间 年月日时分
maybe_rotate(State = #state{fBaseName = Filename, size = Size, fMaxSize = MaxSize,
   ioDevice = OldIoDevice}) when Size > MaxSize ->
   ok = file:close(OldIoDevice),
   ok = file:rename(Filename, Filename ++ ".bak"),
   {ok, NewIoDevice} = file:open(Filename, [write, raw]),
   loop(State#state{size = 0, ioDevice = NewIoDevice});
maybe_rotate(State) ->
   loop(State).

system_continue(_, _, State) ->
   loop(State).

-spec system_terminate(any(), _, _, #state{}) -> no_return().
system_terminate(Reason, _, _, State) ->
   terminate(Reason, State).

system_code_change(Misc, _, _, _) ->
   {ok, Misc}.

-spec terminate(any(), #state{}) -> no_return().
terminate(Reason, #state{ioDevice = IoDevice, buffer = Buffer}) ->
   _ = file:write(IoDevice, lz4f:compress_frame(Buffer)),
   exit(Reason).
